#!/bin/bash
#
#-------------------------------------------------------#
# This file is Juju managed - do not make local changes #
#-------------------------------------------------------#
#
# Copyright 2014 Canonical Ltd.
#
# Check the validity of a mirror's files

set -u
set -e

if [ $# -lt 3 ] ; then
  echo "./check_mirror.sh [scheme] [urlhost] [role] [ipaddrs]"
  echo "e.g., ./check_mirror.sh http archive.ubuntu.com ubuntu 1.2.3.4 4.5.6.7 ..."
  exit 2
fi

scheme=$1 ; shift
urlhost=$1 ; shift
role=$1 ; shift
# remaining args are IPs to check

case $scheme in
    http)
	port=80
	;;
    https)
	port=443
	;;
    *)
	echo "Unknown URL scheme ${scheme} -- cannot determine port! Exiting." >&2
	exit 2
	;;
esac

tmpdir=$(mktemp -d -t check-mirror.${role}.XXXXXXXXXX)
trap "rm -rf -- ${tmpdir}" exit

case ${role} in
    ubuntu)
        unsigned=ubuntu/dists/xenial/Release
        signed=${unsigned}.gpg
        ;;
    ubuntu-cloud-archive)
        unsigned=ubuntu/dists/focal-updates/victoria/Release
        signed=${unsigned}.gpg
        ;;
    releases)
        unsigned=xenial/SHA256SUMS
        signed=${unsigned}.gpg
        ;;
    ports)
        unsigned=dists/xenial/Release
        signed=${unsigned}.gpg
        ;;
    cdimage)
        unsigned=releases/xenial/release/SHA256SUMS
        signed=${unsigned}.gpg
        ;;
    cloud-images)
        unsigned=xenial/current/SHA256SUMS
        signed=${unsigned}.gpg
        ;;
    linuxcontainers)
	unsigned=streams/v1/index.json
	signed=${unsigned}.gpg
	;;
    old-releases)
        unsigned=ubuntu/dists/warty/Release
        signed=${unsigned}.gpg
        ;;
    simple-streams)
        unsigned=juju/images/releases/streams/v1/index.json
        signed=${unsigned}.gpg
        ;;
    maas-images)
	# NOTE(pjdc): ephemeral-v3 is the current stuff and seems to have no fixed paths to signed files we could check
	unsigned=ephemeral-v3/daily/focal/amd64/
	signed=${unsigned}.gpg  # not used
	;;
    *)
        echo "Unknown mirror role, ${role}"
        exit 2
esac

tmpsigned=${tmpdir}/$(basename ${signed})
tmpunsigned=${tmpdir}/$(basename ${unsigned})

case ${role} in
    maas-images|simple-streams)
        for ipaddr in $@; do
            if ! curl -s -o"${tmpunsigned}" --resolve ${urlhost}:${port}:${ipaddr}  "${scheme}://${urlhost}/${unsigned}"; then
                echo "Mirror CRITICAL [${unsigned} failed to fetch from ${ipaddr} for ${scheme}://${urlhost}/]"
                exit 2
            fi
        echo "Mirror OK [${unsigned}]"
        exit 0
        done
        ;;
    cloud-images)
        keyring_file=/usr/share/keyrings/ubuntu-cloudimage-keyring.gpg
        ;;
    linuxcontainers)
	keyring_file=/usr/local/share/ubuntu-mirror-charm.gpg
	;;
    old-releases)
        keyring_file=/usr/share/keyrings/ubuntu-archive-removed-keys.gpg
        ;;
    ubuntu-cloud-archive)
	keyring_file=/usr/share/keyrings/ubuntu-cloud-keyring.gpg
	;;
    simple-streams)
	keyring_file=/usr/local/share/ubuntu-mirror-charm.gpg
	;;
    *)
        keyring_file=/usr/share/keyrings/ubuntu-archive-keyring.gpg
        ;;
esac

for ipaddr in $@; do
    if ! curl -s -o"${tmpsigned}" --resolve "${urlhost}:${port}:${ipaddr}"  "${scheme}://${urlhost}/${signed}"; then
        echo "Mirror CRITICAL [${signed} failed to fetch from ${ipaddr} for ${scheme}://${urlhost}/]"
        exit 2
    fi

    if ! curl -s -o"${tmpunsigned}" --resolve "${urlhost}:${port}:${ipaddr}"  "${scheme}://${urlhost}/${unsigned}"; then
        echo "Mirror CRITICAL [${unsigned} failed to fetch from ${ipaddr} for ${scheme}://${urlhost}/]"
        exit 2
    fi

    output=$(gpgv --keyring ${keyring_file} ${tmpsigned} ${tmpunsigned} 2>&1 || true)
    if ! echo ${output} | fgrep -q "Good"; then
        echo "Mirror CRITICAL [${signed} did not verify correctly for IP ${ipaddr}: $(echo ${output} | head -n 1)]"
        exit 2
    fi
done

echo "Mirror OK [${signed} verified for ${role} at ${scheme}://${urlhost} via IP(s): ${@}]"
exit 0
