#!/bin/bash
#
#-------------------------------------------------------#
# This file is Juju managed - do not make local changes #
#-------------------------------------------------------#
#
# Author: Chris Stratford <chris.stratford@canonical.com>
# Copyright 2014 Canonical Ltd.
#
# Single-pass rsync mirror
#
# ${role}.conf should look like:
# DEST_DIR=/srv/ftp.root/releases
# SOURCE_URL=rsync://releases.ubuntu.com/releases
# RSYNC_PASSWORD=secret

if [ $# -lt 1 ]; then
    myname=$(basename $0)
    echo "Usage: ${myname} (releases|cdimage)"
    exit 1
fi

role=$1
myhostname=$(hostname)
logfile=${HOME}/log/mirror-${role}.log

# Magic to make sure things keep running in the background
# after ssh has gone away
if [ "$2" != "go" ]; then
    annotate-output $0 $1 go > ${logfile} 2>&1 &
    exit 0
fi

set -u

function log {
    echo $1
}

function fatal {
    log $1
    exit 1
}

if [ -f ${HOME}/mirror-config/${role}.conf ]; then
    . ${HOME}/mirror-config/${role}.conf
else
    fatal "${role}.conf file missing - aborting"
fi

export RSYNC_PASSWORD
lockfile="Archive-Update-in-Progress-${role}-${myhostname}"
lockpath="${DEST_DIR}/${lockfile}"

if lockfile -! -l 43200 -r 0 "${lockpath}"; then
    fatal "${myhostname} is unable to start an rsync for ${role}.  Lockfile exists"
fi
trap "rm -f ${lockpath} > /dev/null 2>&1" exit

if [ ! -d ${DEST_DIR} ]; then
    log "${DEST_DIR} does not exist yet, trying to create it..."
    mkdir -p ${DEST_DIR} || fatal "Creation of ${DEST_DIR} failed."
fi

if [ -n "${RSYNC_USER}" ]; then
    url=$(echo ${SOURCE_URL}|sed -e "s,//,//${RSYNC_USER}@,")
else
    url=${SOURCE_URL}
fi

# If upstream is close to us and triggers us via ssh, we may start
# syncing before it has deleted its local lock file, which then
# becomes a "file has vanished" and non-zero exit, causing the trace
# file to not update, which makes the mirror seem stale when it isn't.
# Therefore, let's just:
log "== Sleeping to allow upstream delete its lock file =="
sleep 5

log "== Syncing files from source =="
rsync --verbose --recursive --times --links --hard-links \
  --stats --delete --delete-before \
  --exclude "${lockfile}" --exclude ".trace/${role}-${myhostname}" \
  --exclude ".~tmp~" \
  ${url} ${DEST_DIR} || fatal "Failed to rsync from ${url}."

log "== Making sure permissions are all correct =="
find ${DEST_DIR} -type d ! -perm /o=r,o=x -ls -exec chmod go+rx '{}' +
find ${DEST_DIR} -type f ! -perm /o=r -ls -exec chmod go+r '{}' +

install -d "${DEST_DIR}/.trace/"
date -u > "${DEST_DIR}/.trace/${role}-${myhostname}"

if [ -n "${DOWNSTREAM_MIRRORS}" -a -n "${TRIGGER_KEYFILE}" ]; then
    t=15m
    log "== Triggering downstream mirrors (will give up after $t) =="
    timeout $t ${SCRIPT_DIR}/trigger-downstream-mirrors.sh $role $TRIGGER_KEYFILE $DOWNSTREAM_MIRRORS
fi

savelog ${logfile} > /dev/null 2>&1
rm -f ${lockpath} > /dev/null 2>&1
