#!/bin/bash
#
#-------------------------------------------------------#
# This file is Juju managed - do not make local changes #
#-------------------------------------------------------#
#
# Author: Chris Stratford <chris.stratford@canonical.com>
# Copyright 2014 Canonical Ltd.
#
# Two-pass rsync mirror (for archives)
#
# ${role}.conf should look something like this:
# DEST_DIR=/srv/ftp.root/ubuntu
# SOURCE_URL=rsync://archive.ubuntu.com/ubuntu
# RSYNC_USER=someone
# RSYNC_PASSWORD=secret

if [ $# -lt 1 ]; then
  myname=$(basename $0)
  echo "Usage ${myname} (ubuntu|ports)" 
  exit 1
fi

role=$1
myhostname=$(hostname)
logfile=${HOME}/log/mirror-${role}.log

# Magic to make sure things keep running in the background
# after ssh has gone away
if [ "$2" != "go" ]; then
    annotate-output $0 $1 go > ${logfile} 2>&1 &
    exit 0
fi

set -u

function log {
    echo $1
}

function fatal {
    log $1
    exit 1
}

if [ -f ${HOME}/mirror-config/${role}.conf ]; then
    . ${HOME}/mirror-config/${role}.conf
else
    fatal "${role}.conf file missing - aborting"
fi

export RSYNC_PASSWORD
lockpath="${DEST_DIR}/Archive-Update-in-Progress-${role}-${myhostname}"

if lockfile -! -l 43200 -r 0 "${lockpath}"; then
    fatal "${myhostname} is unable to start an rsync for ${role}.  Lockfile exists"
fi
trap "rm -f ${lockpath} > /dev/null 2>&1" exit

if [ ! -d ${DEST_DIR} ]; then
  log "${DEST_DIR} does not exist yet, trying to create it..."
  mkdir -p ${DEST_DIR} || fatal "Creation of ${DEST_DIR} failed."
fi

if [ -n "${RSYNC_USER}" ]; then
    url=$(echo ${SOURCE_URL}|sed -e "s,//,//${RSYNC_USER}@,")
else
    url=${SOURCE_URL}
fi

# If upstream is close to us and triggers us via ssh, we may start
# syncing before it has deleted its local lock file, which then
# becomes a "file has vanished" and non-zero exit, causing the trace
# file to not update, which makes the mirror seem stale when it isn't.
# Therefore, let's just:
log "== Sleeping to allow upstream delete its lock file =="
sleep 5

log "== Syncing files from source - first stage =="
rsync --verbose --recursive --times --links --hard-links \
  --stats \
  --exclude "Packages*" --exclude "Sources*" \
  --exclude "Release*" --exclude "InRelease" \
  --exclude "Translation*" --exclude "Index" \
  --exclude "${lockpath}" --exclude ".trace/${role}-${myhostname}" \
  --exclude ".~tmp~" \
  ${url} ${DEST_DIR} || fatal "First stage of sync from ${url} failed."

log "== Syncing files from source - second stage =="
rsync --verbose --recursive --times --links --hard-links \
  --stats --delete --delete-after \
  --exclude "${lockpath}" --exclude ".trace/${role}-${myhostname}" \
  --exclude ".~tmp~" \
  ${url} ${DEST_DIR} || fatal "Second stage of sync from ${url} failed."

log "== Making sure permissions are all correct =="
find ${DEST_DIR} -type d ! -perm /o=r,o=x -ls -exec chmod go+rx '{}' +
find ${DEST_DIR} -type f ! -perm /o=r -ls -exec chmod go+r '{}' +

install -d "${DEST_DIR}/.trace/"
date -u > "${DEST_DIR}/.trace/${role}-${myhostname}"

if [ -n "${DOWNSTREAM_MIRRORS}" -a -n "${TRIGGER_KEYFILE}" ]; then
    t=15m
    log "== Triggering downstream mirrors (will give up after $t) =="
    timeout $t ${SCRIPT_DIR}/trigger-downstream-mirrors.sh $role $TRIGGER_KEYFILE $DOWNSTREAM_MIRRORS
fi

savelog ${logfile} > /dev/null 2>&1
rm -f ${lockpath} > /dev/null 2>&1
