#!/usr/bin/python
#
# Copyright 2014 Canonical Ltd.  All rights reserved
# Author: Chris Stratford <chris.stratford@canonical.com>

import json
import base64
import pwd
import os.path
from charmhelpers.core.hookenv import config, log, in_relation_hook, relation_get, relation_type


ROLE_CONFIG_KEYS = [
    "name",
    "aliases",
    "path",
    "description",
    "rsync_log",
    "command",
    "source_url",
    "rsync_auth",
    "rsync_module",
    "sync_time",
    "trigger",
    "apache_late_extra",
    "apache_early_extra",
    "downstream_mirrors",
    "trigger_keyfile",
]


class Config:
    _mirror_map = {}
    _relation_data = {}

    def apache_delete_logs_after(self):
        return int(config("apache_delete_logs_after"))

    def apache_maxclients(self):
        return int(config("apache_maxclients"))

    def apache_maxrequestsperchild(self):
        return int(config("apache_maxrequestsperchild"))

    def apache_maxsparethreads(self):
        return int(config("apache_maxsparethreads"))

    def apache_minsparethreads(self):
        return int(config("apache_minsparethreads"))

    def apache_modules(self):
        module_list = str(config("apache_modules"))
        if not module_list:
            return []
        return module_list.split(" ")

    def apache_serverlimit(self):
        return int(config("apache_serverlimit"))

    def apache_startservers(self):
        return int(config("apache_startservers"))

    def apache_status_allowed_from(self):
        return json.loads(str(config("apache_status_allowed_from")))

    def apache_threadlimit(self):
        return int(config("apache_threadlimit"))

    def apache_threadsperchild(self):
        return int(config("apache_threadsperchild"))

    def app_name(self):
        return str(config("application_name"))

    def archive_logs_after(self):
        return int(config("archive_logs_after"))

    def archive_logs_for(self, service):
        try:
            ret = config("archive_logs_" + service)
        except Exception:
            ret = False
        return ret

    def auth_credentials(self, role, hostname):
        map = self._mirror_map[role]["rsync_auth"]
        user = ""
        password = ""
        if hostname in map:
            user = map[hostname]["user"]
            password = base64.b64decode(map[hostname]["password"])
        elif "<<all>>" in map:
            user = map["<<all>>"]["user"]
            password = base64.b64decode(map["<<all>>"]["password"])
        # No match, assume no auth needed
        return {"user": user, "password": password}

    def config_file(self):
        config_dir = self.mirror_userinfo().pw_dir
        return os.path.join(config_dir, "ubuntu-mirror.conf")

    def disk_inode_crit(self):
        return str(config("disk_inode_crit"))

    def disk_space_crit(self):
        return str(config("disk_space_crit"))

    def disk_space_warn(self):
        return str(config("disk_space_warn"))

    def ftp_delete_logs_after(self):
        return int(config("ftp_delete_logs_after"))

    def ftp_max_instances(self):
        return int(config("ftp_max_instances"))

    def ftp_max_per_source(self):
        return int(config("ftp_max_per_source"))

    def ftp_welcome_banner(self):
        return str(config("ftp_welcome_banner"))

    def logdir(self, service):
        try:
            logdir = str(config(service + "_logdir"))
        except Exception:
            logdir = None
        return logdir

    def mirror_map(self, role):  # noqa: C901
        if role in self._mirror_map:
            return self._mirror_map[role]

        # Some roles are derived from others, which we handle here
        if role == 'ubuntu-unsafe':
            base_role = 'ubuntu'
        else:
            base_role = role

        # Load the map details
        map = {
            'base_role': base_role,
            'unsafe': role.endswith('-unsafe'),
            'tmpl_file': str("apache-{}.tmpl".format(base_role)),
        }
        for key in ROLE_CONFIG_KEYS:
            config_key = "mirror_%s_%s" % (base_role, key)
            try:
                value = config(config_key)
            except ValueError:
                log("CHARM: No config value for {}".format(config_key))
                raise ValueError

            if key in ["trigger", "apache_late_extra", "apache_early_extra"]:
                # Base64 expansion needed
                try:
                    map[key] = str(base64.b64decode(value))
                except TypeError:
                    log("CHARM: Unable to decode {} trigger".format(base_role))
                    map[key] = ""
            elif key in ["aliases", "rsync_auth"]:
                # List/dict expanson needed
                try:
                    map[key] = json.loads(str(value))
                except ValueError:
                    log("CHARM: Unable to decode {} aliases".format(base_role))
                    map[key] = []
            elif key in ["rsync_log"]:
                # No conversion needed
                map[key] = value
            else:
                # Normal string
                map[key] = str(value)

        self._mirror_map[role] = map
        return map

    def mirror_user(self):
        return str(config("mirror_user"))

    def mirror_userinfo(self):
        try:
            mirror_userinfo = pwd.getpwnam(self.mirror_user())
        except KeyError:
            log("CHARM: mirror user '{}' has not been created".format(self.mirror_user()))
            raise KeyError
        return mirror_userinfo

    def nagios_host_context(self):
        return self.relation_data()["nagios_host_context"]

    def nagios_hostname(self):
        return self.relation_data()["nagios_hostname"]

    def nagios_relation_ok(self):
        try:
            if self.nagios_hostname():
                return True
            else:
                return False
        except KeyError:
            return False

    def nagios_servicegroup(self):
        servicegroup = str(config("nagios_servicegroup"))
        if not servicegroup:
            servicegroup = "{}-{}".format(self.nagios_host_context(), self.app_name())
        return servicegroup

    def relation_data(self):
        if not in_relation_hook():
            if self._relation_data:
                return self._relation_data

        # Load any saved data
        try:
            with open(self.config_file(), "r") as f:
                self._relation_data = json.loads(f.read())
        except IOError:
            # File may not be there initially - this is ok
            pass

        if not in_relation_hook():
            return self._relation_data

        log("CHARM: In relation hook {}".format(relation_type()))
        nagios_hostname = relation_get("nagios_hostname")
        if nagios_hostname:
            self._relation_data["nagios_hostname"] = nagios_hostname
        nagios_host_context = relation_get("nagios_host_context")
        if nagios_host_context:
            self._relation_data["nagios_host_context"] = nagios_host_context

        self._relation_data["_This file is auto-generated.  Do not edit_"] = True
        with open(self.config_file(), "w") as f:
            f.write(json.dumps(self._relation_data, sort_keys=True, indent=4))
        return self._relation_data

    # This should contain a list of all roles the charm supports
    def role_list(self):
        return [
            "cdimage",
            "cloud-image",
            "linuxcontainers",
            "maas-images",
            "old-releases",
            "ports",
            "releases",
            "simple-streams",
            "ubuntu",
            "ubuntu-cloud-archive",
            "ubuntu-unsafe",
        ]

    def roles(self):
        if not config("role_map"):
            # No roles defined yet - this is ok
            return {}
        try:
            roles = json.loads(str(config("role_map")))
        except ValueError:
            print "Error found"
            raise ValueError("Error decoding roles JSON")
        return roles

    def rsync_max_connections(self):
        return int(config("rsync_max_connections"))

    def rsync_max_connections_per_source(self):
        return int(config("rsync_max_connections_per_source"))

    def rsync_motd(self):
        try:
            motd = str(base64.b64decode(config("rsync_motd")))
        except TypeError:
            motd = ""
        return motd

    def rsync_motdfile(self):
        return str(config("rsync_motdfile"))

    def rsyncd_dir(self):
        return "/etc/rsync-juju.d"

    def rsync_use_xinetd(self):
        return config("rsync_use_xinetd")

    def script_dir(self):
        return str(config("script_dir"))

    def package_status(self):
        return str(config("package_status"))
