#!/usr/bin/python
#
# Copyright 2014 Canonical Ltd.  All rights reserved
# Author: Chris Stratford <chris.stratford@canonical.com>

import base64
import datetime
import errno
import json
import os
import os.path
import platform
import pwd
import shutil
import socket
import subprocess
import sys
import types
import yaml

from glob import glob

from charmhelpers.core.host import (
    adduser,
    lsb_release,
    mkdir,
    service_reload,
    service_restart,
    user_exists,
    write_file,
)
from charmhelpers.core.hookenv import (
    Hooks,
    charm_dir,
    is_relation_made,
    local_unit,
    log,
    relation_id,
    relation_ids,
    relation_set,
    status_set,
)
from charmhelpers.core import sysctl
from charmhelpers.fetch import (
    apt_install,
    get_upstream_version,
)
from charmhelpers.payload.execd import execd_run
from Cheetah.Template import Template
from Config import Config

hooks = Hooks()

apache_tls_settings = {
    'ssl_cipher_suite': 'EECDH+AESGCM+AES128:EDH+AESGCM+AES128:EECDH+AES128:EDH+AES128:ECDH+AESGCM+AES128:aRSA+AESGCM+AES128:ECDH+AES128:DH+AES128:aRSA+AES128:EECDH+AESGCM:EDH+AESGCM:EECDH:EDH:ECDH+AESGCM:aRSA+AESGCM:ECDH:DH:aRSA:HIGH:!MEDIUM:!aNULL:!NULL:!LOW:!3DES:!DSS:!EXP:!PSK:!SRP',  # noqa: E501
    'ssl_honor_cipher_order': 'on',
    # per IS policy, TLS 1.2 is the minimum acceptable version
    'ssl_protocol': 'ALL -SSLv2 -SSLv3 -TLSv1 -TLSv1.1',
}

nagios_etcdir = "/etc/nagios/nrpe.d"
nagios_exportdir = "/var/lib/nagios/export"

required_pkgs = [
    'apache2',
    'curl',         # for check-mirror.sh
    'devscripts',   # provides annotate-output
    'geoip-database',
    'libapache2-mod-geoip',
    'logrotate',
    'procmail',     # provides lockfile
    'rsync',
    'ubuntu-cloud-keyring',     # for check-mirror.sh
    'vsftpd',
    'xinetd',
]

service_affecting_packages = ['apache2']

apache_modules = [
    'expires',
    'geoip',
    'headers',
    'rewrite',
]

scripts_to_copy = [
    'mirror-1stage.sh',
    'mirror-2stage.sh',
    'mirror-linuxcontainers.sh',
    'check-mirror.sh',
    'check-updates.sh',
]

bind_mount_comment = "# Bind mount managed by the ubuntu-mirror charm"


def juju_header():
    header = ("#-------------------------------------------------#\n"
              "# This file is Juju managed - do not edit by hand #\n"
              "#-------------------------------------------------#\n")
    return header


# Google Public DNS addresses are good "cannot possibly be internal" destinations
GET_DEFAULT_ADDRESS_DESTINATIONS = {
    socket.AF_INET: ("8.8.8.8", 53),
    socket.AF_INET6: ("2001:4860:4860::8888", 53),
}


def ensure_symlink(target, link_path):
    """Ensure that link_path is a symlink to target.  Do not follow link_path.
    Raise an exception if link_path exists but is not a symlink."""
    need_unlink = False
    try:
        current = os.readlink(link_path)
        if current == target:
            log("Skipping {} - symlink exists and already points to {}".format(link_path, current))
            return
        need_unlink = True
    except OSError as ose:
        if ose.errno != errno.ENOENT:
            raise
    except Exception:
        raise

    # There's no non-racy way to replace a symlink.  Even `ls -sfn target link_path` does this.
    if need_unlink:
        os.unlink(link_path)
    os.symlink(target, link_path)
    log("Symlink {} at {}, pointing to {}".format("replaced" if need_unlink else "created", link_path, target))


def get_default_address(af):
    """Determine our default address for the given address family."""
    try:
        s = socket.socket(af, socket.SOCK_DGRAM)
        s.connect(GET_DEFAULT_ADDRESS_DESTINATIONS[af])
        return s.getsockname()[0]
    except Exception as e:
        log("CHARM: Unable to determine local address for address family {}: {}".format(af, str(e)))
        return None
    finally:
        s.close()


def ensure_package_status(packages, status):
    if status in ['install', 'hold']:
        selections = ''.join(['{} {}\n'.format(package, status)
                              for package in packages])
        dpkg = subprocess.Popen(['dpkg', '--set-selections'],
                                stdin=subprocess.PIPE)
        dpkg.communicate(input=selections)


def check_call(cmd):
    cmdline = " ".join(cmd)
    log("CHARM: Calling '{}'".format(cmdline))
    try:
        subprocess.check_call(cmd)
        return 0
    except subprocess.CalledProcessError as e:
        log("CHARM: '{}' failed. Error: {}".format(cmdline, e))
        return e.returncode


def check_output(cmd):
    cmdline = " ".join(cmd)
    log("CHARM: Calling '{}'".format(cmdline))
    try:
        output = subprocess.check_output(cmd)
        return output.strip()
    except subprocess.CalledProcessError as e:
        log("CHARM: '{}' failed. Error: {}".format(cmdline, e))
        return None


def get_managed_mounts():
    managed_mounts = []
    with open("/etc/fstab", "r") as f:
        for line in f:
            if line.strip().endswith(bind_mount_comment):
                managed_mounts.append(line.replace(bind_mount_comment, '').strip())
    return managed_mounts


def add_bind_mount(fstab_line):
    target = fstab_line.split()[1]
    line_to_write = "{} {}\n".format(fstab_line, bind_mount_comment)
    fstab_path = "/etc/fstab"

    suffix = "_{:%Y-%m-%d_%H:%M:%S}".format(datetime.datetime.now())
    tempfile_suffix = "_temp" + suffix
    backup_suffix = "_backup" + suffix

    tempfile_path = fstab_path + tempfile_suffix
    backup_path = fstab_path + backup_suffix

    shutil.copyfile(fstab_path, tempfile_path)
    shutil.copyfile(fstab_path, backup_path)

    with open(tempfile_path, "a") as f:
        f.write(line_to_write)
    os.rename(tempfile_path, fstab_path)
    if not os.path.isdir(target):
        mkdir(target)
    check_call("mount {}".format(target).split())


def remove_bind_mount(fstab_line):
    target = fstab_line.split()[1]
    removed = False
    fstab_path = "/etc/fstab"

    suffix = "_{:%Y-%m-%d_%H:%M:%S}".format(datetime.datetime.now())
    tempfile_suffix = "_temp" + suffix
    backup_suffix = "_backup" + suffix

    tempfile_path = fstab_path + tempfile_suffix
    backup_path = fstab_path + backup_suffix

    shutil.copyfile(fstab_path, backup_path)

    with open(fstab_path, "r") as f:
        lines = f.readlines()
    with open(tempfile_path, "w") as f:
        for line in lines:
            if line.strip().endswith(bind_mount_comment):
                if line.replace(bind_mount_comment, '').strip() == fstab_line:
                    removed = True
                    continue
            f.write(line)
    os.rename(tempfile_path, fstab_path)
    if removed:
        check_call("umount {}".format(target).split())
        check_call("rmdir {}".format(target).split())


def update_bind_mounts(required_mounts):
    managed_mounts = get_managed_mounts()

    for required_mount in required_mounts:
        if required_mount not in managed_mounts:
            add_bind_mount(required_mount)

    for managed_mount in managed_mounts:
        if managed_mount not in required_mounts:
            remove_bind_mount(managed_mount)


def configure_bind_mounts(conf, hostname):
    roles = conf.roles()
    paths = []
    required_mounts = []
    ftp_root = "/srv/ftp.root"  # configure_vsftp() uses this hard-coded path, so we will too
    if roles.get(hostname):
        for role in roles.get(hostname):
            try:
                mirror = conf.mirror_map(role)
            except ValueError:
                log("CHARM: Missing details - skipping {}".format(role))
                continue

            paths.append(mirror['path'])
    else:
        print("DEBUG: configure_bind_mounts found no defined roles for host {}".format(hostname))

    if not paths:
        print("DEBUG: configure_bind_mounts found no configured paths for host {}".format(hostname))

    for path in paths:
        if not path.startswith(ftp_root):
            dir_name = os.path.basename(os.path.normpath(path))
            target = os.path.join("/srv/ftp.root", dir_name)
            print("DEBUG: configure_bind_mounts needs to bind-mount {} to {}".format(path, target))
            fstab_line = "{} {} none defaults,bind,nofail 0 0".format(path, target)
            required_mounts.append(fstab_line)

    update_bind_mounts(required_mounts)


def configure_scripts(conf, hostname):
    mkdir(conf.script_dir())
    for script in scripts_to_copy:
        src = os.path.join(charm_dir(), "files", script)
        dest = os.path.join(conf.script_dir(), script)
        shutil.copyfile(src, dest)
        os.chmod(dest, 0o755)


def file_from_template(tmpl, dest, searchList, add_header=True):
    template_file = os.path.join(charm_dir(), "templates", tmpl)
    t = Template(file=template_file, searchList=searchList)
    contents = ''
    if add_header:
        contents += juju_header()
    contents += str(t)
    write_file(dest, contents, perms=0o444)
    log("CHARM: Installed template {} to {}".format(tmpl, dest))


def get_ssh_keyfile(username):
    keyfile = None
    with open("/etc/ssh/sshd_config", "r") as f:
        for line in f:
            if line.startswith("AuthorizedKeysFile"):
                # Assume the first one is the one we want
                keyfile = line.split()[1]
    if not keyfile:
        # Assume default
        keyfile = "%h/.ssh/authorized_keys"

    userinfo = pwd.getpwnam(username)
    newkeyfile = keyfile.replace("%%", "%")
    newkeyfile = newkeyfile.replace("%h", userinfo[5])
    newkeyfile = newkeyfile.replace("%u", username)

    return newkeyfile


def has_roles(conf, hostname):
    """Returns true if there are specific roles for `hostname` hostname
    or global roles defined for <<all>> hostnames"""
    roles = conf.roles()
    if hostname in roles or '<<all>>' in roles:
        return True
    return False


def get_roles(conf, hostname):
    """Return the list of specific roles for `hostname` hostname.
    If there are none, but global roles are defined for <<all>> hostname,
    return these instead of an empty list."""
    if not has_roles(conf, hostname):
        return []

    roles = conf.roles()
    if hostname in roles:
        return roles[hostname]
    else:
        return roles['<<all>>']


#
# Set up cronned rsyncing to pull archive contents from masters
# Creation of any necessary destination directories is the responsibility
# of the sync script specified in ${mirror}_command
#
def configure_rsync_client(conf, hostname):  # noqa: C901
    log("CHARM: Configuring rsync client")
    mirror_userinfo = conf.mirror_userinfo()
    mirror_confdir = os.path.join(mirror_userinfo.pw_dir, "mirror-config")
    mirror_logdir = os.path.join(mirror_userinfo.pw_dir, "log")

    if not os.path.isdir(mirror_confdir):
        mkdir(mirror_confdir)

    if not os.path.isdir(mirror_logdir):
        mkdir(mirror_logdir)
        os.chmod(mirror_logdir, 0o755)
        os.chown(mirror_logdir, mirror_userinfo.pw_uid, mirror_userinfo.pw_gid)

    if not has_roles(conf, hostname):
        log("CHARM: hostname {} not found in role_map".format(hostname))

    enabled_configs = [f for f in os.listdir(mirror_confdir) if os.path.isfile(os.path.join(mirror_confdir, f))]
    wanted_configs = []

    cron = juju_header()
    for role in get_roles(conf, hostname):
        try:
            mirror = conf.mirror_map(role)
            if mirror["unsafe"]:
                continue
            config_file = "%s.conf" % role
            wanted_configs.append(config_file)
        except ValueError:
            log("CHARM: Missing details - skipping {}".format(role))
        else:
            auth_credentials = conf.auth_credentials(role, hostname)
            tmpl_data = {
                "source_url": mirror["source_url"],
                "dest_dir": mirror["path"],
                "user": auth_credentials["user"],
                "password": auth_credentials["password"],
                "script_dir": conf.script_dir(),
                "downstream_mirrors": mirror["downstream_mirrors"],
                "trigger_keyfile": mirror["trigger_keyfile"],
            }
            config_path = os.path.join(mirror_confdir, config_file)
            file_from_template("mirror-conf.tmpl", config_path, tmpl_data)
            os.chmod(config_path, 0o440)
            os.chown(config_path, mirror_userinfo.pw_uid, mirror_userinfo.pw_gid)

            sync_time = mirror["sync_time"]
            if sync_time != 'ondemand':
                sync_cmd = mirror["command"]
                line = "%s %s %s/%s %s\n" % (sync_time, conf.mirror_user(), conf.script_dir(), sync_cmd, role)
                cron += line

    # Remove unwanted config files
    unwanted_configs = list(set(enabled_configs) - set(wanted_configs))
    for unwanted in unwanted_configs:
        try:
            config_file = os.path.join(mirror_confdir, unwanted)
            os.unlink(config_file)
        except OSError:
            log("CHARM: Unable to delete config {}".format(config_file))

    write_file("/etc/cron.d/ubuntu-mirror", cron)
    log("CHARM: Finished configuring rsync client")


#
# Set up rsync server
#
# FIXME: We rely on basenode to provide rsyncd, but overwrite its
# config with our own to give connection limits and motd.
# Change to update the file rather than overwriting.
#
def configure_rsync_server(conf, hostname):
    log("CHARM: Configuring rsync server")

    if not has_roles(conf, hostname):
        log("CHARM: hostname {} not found in role_map - not configuring rsync".format(hostname))
        return

    targets = {}
    for role in get_roles(conf, hostname):
        try:
            mirror = conf.mirror_map(role)
            if mirror["unsafe"]:
                continue
        except ValueError:
            log("CHARM: Missing details - skipping {}".format(role))
            continue
        else:
            targets[role] = mirror

    motdfile = conf.rsync_motdfile()
    tmpl_data = {}
    tmpl_data["max_connections"] = conf.rsync_max_connections()
    tmpl_data["max_connections_per_source"] = conf.rsync_max_connections_per_source()
    tmpl_data["motdfile"] = motdfile
    tmpl_data["logdir"] = conf.logdir("rsync")
    tmpl_data["targets"] = targets

    # Write the motd file
    mkdir(os.path.dirname(motdfile))
    write_file(motdfile, conf.rsync_motd(), perms=0o444)

    # Make sure the logdir exists
    mkdir(tmpl_data["logdir"])

    # Make sure the rsync dir exists
    mkdir(conf.rsyncd_dir())

    file_from_template("rsync-base-cfg.tmpl", "/etc/rsyncd.conf", tmpl_data)
    file_from_template("rsync-mirrors-cfg.tmpl", os.path.join(conf.rsyncd_dir(), "010-ubuntu-mirror.conf"), tmpl_data)

    # xinetd is opt-in, now.
    if conf.rsync_use_xinetd():
        file_from_template("rsync-default.tmpl", "/etc/default/rsync", tmpl_data)
        file_from_template("rsync-xinetd.tmpl", "/etc/xinetd.d/rsync", tmpl_data)
        check_call('systemctl disable rsyncd.socket'.split())
        check_call('systemctl stop rsyncd.socket'.split())
    # Need to tell systemd to enable rsync and then start it
    # but only in 15.04 or greater per https://wiki.ubuntu.com/SystemdForUpstartUsers
    elif float(platform.linux_distribution()[1]) >= 15.04:
        # We disable rsync and run rsync over systemd sockets to allow limiting
        # connections per source/IP
        # We clean up any xinetd files as well.
        try:
            os.unlink("/etc/default/rsync")
            os.unlink("/etc/xinetd.d/rsync")
        except OSError:
            pass  # If the files are already gone, we're happy.
        check_call('systemctl disable rsync'.split())
        check_call('systemctl stop rsync'.split())

        log("CHARM: Installing rsyncd system socket service files")
        src = os.path.join(charm_dir(), "files", "rsyncd-systemd.service")
        shutil.copyfile(src, "/etc/systemd/system/rsyncd@.service")
        file_from_template("rsyncd-systemd-socket.tmpl", "/etc/systemd/system/rsyncd.socket", tmpl_data)
        log("CHARM: Starting up / reloading rsyncd.socket")
        check_call('systemctl daemon-reload'.split())
        check_call('systemctl enable rsyncd.socket'.split())
        check_call('systemctl start rsyncd.socket'.split())
    # pre-15.04 installations will need to explicitly turn on xinetd,
    # or continue with previous upstart behaviours.
    check_call('systemctl restart xinetd.service'.split())

    log("CHARM: Finished configuring rsync server")


def get_role_config(host_roles, role):
    """Get the role configuration for `role` as defined in host_roles
    or return an empty dict if not defined"""
    if isinstance(host_roles, types.DictType) and host_roles[role]:
        return host_roles[role]
    return {}


#
# Set up apache config files - adding or removing vhosts as needed
#
def configure_apache(conf, hostname):  # noqa: C901
    log("CHARM: Configuring apache")
    ensure_package_status(service_affecting_packages, conf.package_status())
    apache_dir = "/etc/apache2"
    enabled_dir = os.path.join(apache_dir, "sites-enabled")
    available_dir = os.path.join(apache_dir, "sites-available")
    apache_logdir = conf.logdir("apache")
    ubuntu_rel = lsb_release()['DISTRIB_CODENAME']

    # Ensure the apache logdir exists
    mkdir(apache_logdir)
    os.chmod(apache_logdir, 0o755)

    if not has_roles(conf, hostname):
        log("CHARM: hostname {} not found in role_map - not configuring apache".format(hostname))
        return

    wanted_vhosts = []
    all_addresses = set()
    all_ports = set()
    host_roles = get_roles(conf, hostname)

    # Create vhost configs for all active roles
    for role in host_roles:
        try:
            mirror = conf.mirror_map(role)
        except ValueError:
            log("CHARM: Missing details - skipping {}".format(role))
        else:
            role_config = get_role_config(host_roles, role)
            if mirror['unsafe'] and 'addresses' not in role_config:
                log("CHARM: {} has no addresses defined - not configuring".format(role))
                continue
            wanted_vhosts.append(role)
            sites_available = os.path.join(available_dir, role + ".conf")
            sites_enabled = os.path.join(enabled_dir, role + ".conf")
            tmpl_data = mirror
            tmpl_data.update(apache_tls_settings)
            if 'name' in role_config:
                tmpl_data['name'] = role_config['name']
                # If name is overridden then aliases must be too:
                tmpl_data['aliases'] = role_config.get('aliases', [])
            tmpl_data["logdir"] = apache_logdir
            tmpl_data["addresses"] = role_config.get('addresses', ['*'])
            all_addresses.update(tmpl_data["addresses"])

            ports = [80]
            if role_config.get('https'):
                ports.append(443)
                if role == 'linuxcontainers':
                    ports.append(8443)
            all_ports.update(ports)
            tmpl_data["ports"] = ports

            file_from_template(mirror['tmpl_file'], sites_available, tmpl_data)
            ensure_symlink(sites_available, sites_enabled)

            # And make sure the docroot exists
            if role == "ubuntu":
                docroot = os.path.join("/srv", mirror["name"], "www")
                linkdest = os.path.join(docroot, "ubuntu")
            else:
                docroot = os.path.join("/srv", mirror["name"])
                linkdest = os.path.join(docroot, "www")

            # Ensure the directories exist and have the correct perms
            mirror_userinfo = conf.mirror_userinfo()
            mkdir(docroot)
            mkdir(mirror["path"])
            os.chmod(mirror["path"], 0o755)
            os.chown(mirror["path"], mirror_userinfo.pw_uid, mirror_userinfo.pw_gid)
            ensure_symlink(mirror["path"], linkdest)

    # archive.ubuntu.com must not open port 443.  Three factors align
    # to prevent this: 1) we do not set "addresses" or "https" in the
    # role_map; 2) the template doesn't make non-port-80 sockets for
    # members of "addresses"; 3) the wildcard address is discarded.
    all_addresses.discard('*')
    file_from_template('apache-listen-ports.conf.tmpl', '/etc/apache2/ports.conf',
                       {'addresses': sorted(all_addresses), 'ports': sorted(all_ports)})

    # Enable ssl module, if not already enabled
    restart_for_ssl = False
    if 443 in all_ports:
        if check_call(['/usr/sbin/a2query', '-m', 'ssl']) != 0:
            check_call(['/usr/sbin/a2enmod', 'ssl'])
            restart_for_ssl = True

    # Remove all other vhosts
    unwanted_vhosts = set(conf.role_list()) - set(wanted_vhosts)
    # always disable default
    default_site = "000-default.conf"
    if os.path.exists(os.path.join(enabled_dir, default_site)):
        unwanted_vhosts.add(default_site)
    for vhost in unwanted_vhosts:
        check_call(["/usr/sbin/a2dissite", vhost])
        if os.path.isfile(os.path.join(available_dir, vhost)):
            try:
                os.unlink(os.path.join(available_dir, vhost))
            except OSError:
                log("CHARM: Unable to delete vhost {}".format(vhost))

    # Set up server limits
    restart_for_mpm = configure_apache_mpm(conf, hostname)

    # Configure status module
    tmpl_data = {}
    tmpl_data["allowed"] = conf.apache_status_allowed_from()
    tmpl_file = "apache-status.tmpl"
    file_from_template(tmpl_file, "/etc/apache2/mods-available/status.conf", tmpl_data)
    check_call(["/usr/sbin/a2enmod", "status"])

    # Enable any extra modules
    for module in conf.apache_modules():
        check_call(["/usr/sbin/a2enmod", module])

    # Set up log rotation
    tmpl_data = {}
    tmpl_data["logdir"] = apache_logdir
    tmpl_data["delete_after"] = conf.apache_delete_logs_after()
    file_from_template("apache-logrotate.tmpl", "/etc/logrotate.d/apache2", tmpl_data)

    # Set up systemd resource limits
    if ubuntu_rel not in ['precise', 'trusty']:
        systemd_apachedir = '/etc/systemd/system/apache2.service.d'
        tmpl_data = {}
        tmpl_data["max_fds"] = 65535
        if not os.path.isdir(systemd_apachedir):
            mkdir(systemd_apachedir)
        override_file = '{}/override.conf'.format(systemd_apachedir)
        file_from_template("systemd-apache-override.tmpl", override_file, tmpl_data)

    if restart_for_mpm or restart_for_ssl:
        log("CHARM: Restarting apache")
        service_restart("apache2")
    else:
        log("CHARM: Reloading apache")
        service_reload("apache2")

    # Notify any interested parties
    configure_extra_mirrors(conf, hostname)
    log("CHARM: Finished configuring apache")


# Set up the mpm config
# 2.4 defaults to event
# Returns True is something changed, False otherwise
def configure_apache_mpm(conf, hostname):  # noqa: C901
    log("CHARM: Configuring mpm")
    restart_required = False
    apache_version = get_upstream_version("apache2")[0:3]
    if apache_version != "2.4":
        error_message = "Apache version {} is not supported!".format(apache_version)
        log("CHARM: " + error_message)
        status_set('blocked', error_message)
        sys.exit(1)

    # Revert back to MPM worker for now:
    #   https://bz.apache.org/bugzilla/show_bug.cgi?id=53555
    # mpm_file = "/etc/apache2/mods-available/mpm_event.conf"
    # mpm_module = "mpm_event"
    mpm_file = "/etc/apache2/mods-available/mpm_worker.conf"
    mpm_module = "mpm_worker"
    current_mpm_module = 'mpm_{}'.format(check_output(["/usr/sbin/a2query", "-M"]))
    if current_mpm_module == 'mpm_None':
        log("CHARM: Query for existing module failed, setting to {}"
            .format(mpm_module))
        check_call(["/usr/sbin/a2enmod", mpm_module])
        restart_required = True
    elif current_mpm_module != mpm_module:
        log("CHARM: Switching MPM modules from {} to {}"
            .format(current_mpm_module, mpm_module))
        check_call(["/usr/sbin/a2dismod", current_mpm_module])
        check_call(["/usr/sbin/a2enmod", mpm_module])
        restart_required = True

    # Get existing values; we'll check None to determine
    # whether a value is set in the config file or not.
    data = {"StartServers": None, "MinSpareThreads": None, "MaxSpareThreads": None,
            "ThreadLimit": None, "ThreadsPerChild": None, "ServerLimit": None,
            "MaxClients": None, "MaxRequestWorkers": None,
            "MaxRequestsPerChild": None, "MaxConnectionsPerChild": None}
    if os.path.isfile(mpm_file):
        with open(mpm_file, "r") as f:
            for line in f:
                s = line.split()
                if not s:
                    continue
                if s[0] in data:
                    data[s[0]] = int(s[1])

    # Check if anything has changed
    new_data = {}
    new_data["StartServers"] = conf.apache_startservers()
    new_data["MinSpareThreads"] = conf.apache_minsparethreads()
    new_data["MaxSpareThreads"] = conf.apache_maxsparethreads()
    new_data["ThreadLimit"] = conf.apache_threadlimit()
    new_data["ThreadsPerChild"] = conf.apache_threadsperchild()
    new_data["ServerLimit"] = conf.apache_serverlimit()
    # 2.2 -> 2.4 variable name changes
    new_data["MaxRequestWorkers"] = conf.apache_maxclients()
    new_data["MaxConnectionsPerChild"] = conf.apache_maxrequestsperchild()
    new_data["MPM_Module"] = mpm_module

    update_mpm_config = False

    restart_required_keys = [
        'ServerLimit',
        'ThreadLimit',
    ]

    for d in data:
        if data[d] is None:
            # not specified in live config, so skip it
            continue
        if new_data[d] != data[d]:
            log("CHARM: {} changed: {} != {}".format(
                d, new_data[d], data[d]))
            update_mpm_config = True
            if d in restart_required_keys:
                restart_required = True

    if update_mpm_config:
        tmpl_data = new_data
        tmpl_file = "apache-mpm.tmpl"
        file_from_template(tmpl_file, mpm_file, tmpl_data)

    return restart_required


#
# Set up vsftp
#
def configure_vsftp(conf, hostname):
    log("CHARM: Configuring vsftp")

    # Set the ftp user's homedir where we want it
    check_call('/usr/sbin/usermod -d /srv/ftp.root ftp'.split())

    # Make sure the ftplog dir exists
    mkdir(conf.logdir("ftp"))

    tmpl_data = {}
    tmpl_data["max_instances"] = conf.ftp_max_instances()
    tmpl_data["max_per_source"] = conf.ftp_max_per_source()
    tmpl_data["logdir"] = conf.logdir("ftp")
    tmpl_data["welcome_banner"] = conf.ftp_welcome_banner()
    tmpl_data["delete_after"] = conf.ftp_delete_logs_after()
    file_from_template("vsftp-xinetd.tmpl", "/etc/xinetd.d/ftp", tmpl_data)
    file_from_template("vsftp-conf.tmpl", "/etc/vsftpd.conf", tmpl_data)
    file_from_template("vsftp-logrotate.tmpl", "/etc/logrotate.d/vsftpd", tmpl_data)
    log("CHARM: Finished configuring vsftp")


SSH_TRIGGER_TEMPLATE = 'no-port-forwarding,no-X11-forwarding,no-agent-forwarding,no-pty,command="{command}",{trigger}'


# cdimage-builder.internal triggers releases and cdimage using the
# same key, so when this unit has both roles, sync both of them.
def merge_shared_key_triggers(conf, triggers):
    """Merge (in place) triggers that use the same ssh key."""

    cdimage_key = extract_ssh_public_key(triggers.get('cdimage', ''))
    releases_key = extract_ssh_public_key(triggers.get('releases', ''))
    if cdimage_key and releases_key and cdimage_key == releases_key:
        releases = conf.mirror_map('releases')
        cdimage = conf.mirror_map('cdimage')
        command = '( {releases_command} releases ; {cdimage_command} cdimage ) &'.format(
            releases_command=os.path.join(conf.script_dir(), releases['command']),
            cdimage_command=os.path.join(conf.script_dir(), cdimage['command']),
        )
        triggers['releases'] = SSH_TRIGGER_TEMPLATE.format(
            command=command,
            trigger=releases['trigger'],
        )
        del(triggers['cdimage'])


def extract_ssh_public_key(key):
    """Return the field after the field that begins with "ssh-", which
    should be the key itself.  This function can therefore cope with
    fairly arbitrarily-decorated ssh keys, e.g. those with "from", etc."""

    kelts = key.split(' ')
    while kelts:
        # This doesn't work for every type of key, but nobody should be using ECDSA anyway.
        if kelts.pop(0).startswith('ssh-'):
            return kelts[0]
    return None


def make_triggers(conf, hostname):
    """Return this unit's triggers in a dict, keyed by role."""

    if not has_roles(conf, hostname):
        return {}

    triggers = {}
    for role in get_roles(conf, hostname):
        try:
            mirror = conf.mirror_map(role)
            if mirror["unsafe"]:
                continue
        except ValueError:
            log("CHARM: Missing details - skipping {}".format(role))
            continue
        else:
            if mirror["trigger"]:
                triggers[role] = SSH_TRIGGER_TEMPLATE.format(
                    command="%s %s &" % (os.path.join(conf.script_dir(), mirror["command"]), role),
                    trigger=mirror["trigger"],
                )

    return triggers


#
# Set up any external ssh sync triggers
#
def configure_triggers(conf, hostname):
    if not has_roles(conf, hostname):
        return

    triggers = make_triggers(conf, hostname)
    merge_shared_key_triggers(conf, triggers)

    keyfile = get_ssh_keyfile(conf.mirror_user())
    ssh_dir = os.path.dirname(keyfile)
    if not os.path.isdir(ssh_dir):
        mkdir(ssh_dir)
    write_file(keyfile, '\n'.join(triggers.values()) + '\n')


def mountpoint_of(path):
    """Test path, and its parent directories, returning the first mount point encountered."""
    path = os.path.realpath(path)

    while True:
        if os.path.ismount(path):
            return path
        if path == '/':
            break
        path = os.path.realpath(os.path.join(path, '..'))

    # If we reach / and / isn't a mount point, something's wrong somewhere.
    raise Exception("Terminated without finding a mount point")


def collect_mountpoints(paths):
    mountpoints = []
    for path in paths:
        mountpoints.append(mountpoint_of(path))
    return set(mountpoints)


def generate_disk_check(conf, path):
    mount_name = path.replace('/', '_').strip('_')

    file_from_template(
        "nagios-check_disk.tmpl",
        os.path.join(nagios_etcdir, "check_disk_{mount_name}.cfg".format(mount_name=mount_name)),
        {
            "mount_name": mount_name,
            "space_warn": conf.disk_space_warn(),
            "space_crit": conf.disk_space_crit(),
            "inode_crit": conf.disk_inode_crit(),
            "path": path,
        }
    )
    file_from_template(
        "nagios-config.tmpl",
        os.path.join(nagios_exportdir, "service__{hostname}_disk_{mount_name}.cfg".format(
            hostname=conf.nagios_hostname(), mount_name=mount_name)),
        {
            "use": "active-service",
            "nagios_hostname": conf.nagios_hostname(),
            "description": "Disk space on {}".format(path),
            "check_command": "check_nrpe!check_disk_{}".format(mount_name),
            "servicegroup": conf.nagios_servicegroup(),
        }
    )
    log("CHARM: Adding disk check for {}".format(path))


def generate_global_rsync_check(conf, hostname, host_roles):
    # One extra for the rsync daemon itself, another two
    # per role for pulling from syncproxy, one more for luck.
    rsync_crit = conf.rsync_max_connections() + 1 + len(host_roles) * 2 + 1
    rsync_warn = conf.rsync_max_connections() + 1  # all client slots used

    file_from_template(
        "nagios-check_rsync.tmpl",
        os.path.join(nagios_etcdir, "check_global_rsync.cfg"),
        {
            "name": "global",
            "rsync_crit": rsync_crit,
            "rsync_warn": rsync_warn,
        }
    )
    file_from_template(
        "nagios-config.tmpl",
        os.path.join(nagios_exportdir, "service__{hostname}_global_rsync.cfg".format(
            hostname=conf.nagios_hostname())),
        {
            "use": "active-service",
            "nagios_hostname": conf.nagios_hostname(),
            "description": "rsync processes on {}".format(hostname),
            "check_command": "check_nrpe!check_rsync_global",
            "servicegroup": conf.nagios_servicegroup(),
        }
    )
    log("CHARM: Adding global rsync check")


def configure_nrpe(conf, hostname):  # noqa: C901
    if not conf.nagios_relation_ok():
        log("CHARM: Nagios relation not established - not configuring nrpe")
        return

    role_checks = {
        "check_virthost": {"desc": "Check apache virthost"},
        "check_mirror": {"desc": "GPG signature via HTTP"},
        "check_mirror_https": {
            "desc": "GPG signature via HTTPS",
            "when": "https",
        },
        "check_updates": {"desc": "Mirror update status"},
    }

    # No nagios dirs probably means no nagios, so don't bother with the rest
    if not (os.path.isdir(nagios_etcdir) and os.path.isdir(nagios_exportdir)):
        log("CHARM: Essential nagios dirs missing - skipping Nagios checks")
        return

    # Add Nagios role checks
    wanted_checks = []
    paths = []

    if has_roles(conf, hostname):
        host_roles = get_roles(conf, hostname)
        # We generate a single check here, and delete the legacy per-role checks later.
        generate_global_rsync_check(conf, hostname, host_roles)
        tmpl_data = {}
        for role in host_roles:
            try:
                mirror = conf.mirror_map(role)
            except ValueError:
                log("CHARM: Missing details - skipping {}".format(role))
                continue

            paths.append(mirror['path'])
            tmpl_data["name"] = mirror["name"]
            role_config = get_role_config(host_roles, role)
            addresses = role_config.get('addresses', [])
            if not addresses:
                ipaddr = get_default_address(socket.AF_INET)
                if not ipaddr:
                    error = 'Could not determine default IP address!'
                    status_set('blocked', error)
                    raise Exception(error)
                addresses.append(ipaddr)
                # IPv6 can't be a requirement, and get_default_address
                # logs the problem for us, so we just proceed silently.
                ip6addr = get_default_address(socket.AF_INET6)
                if ip6addr:
                    addresses.append("[{}]".format(ip6addr))
            tmpl_data["ipaddr"] = ' '.join(addresses)
            tmpl_data["port"] = 80
            tmpl_data["role"] = mirror["base_role"]
            tmpl_data["path"] = mirror["path"]
            tmpl_data["hostname"] = hostname
            tmpl_data["use"] = "active-service"
            tmpl_data["nagios_hostname"] = conf.nagios_hostname()
            if role == "linuxcontainers":
                tmpl_data["tracepath"] = os.path.join(mirror["path"], ".serial")
            else:
                tmpl_data["tracepath"] = os.path.join(mirror["path"], ".trace", mirror["base_role"] + "-" + hostname)

            for check, check_details in role_checks.items():
                if check_details.get('when') and not role_config.get(check_details['when']):
                    continue
                # Nagios check file
                checkfile = "%s_%s.cfg" % (check, mirror["name"])
                checkpath = os.path.join(nagios_etcdir, checkfile)
                tmplfile = "nagios-%s.tmpl" % check
                file_from_template(tmplfile, checkpath, tmpl_data)
                wanted_checks.append(checkfile)

                # Nagios config file
                tmpl_data["description"] = "%s for %s on %s" % (check_details["desc"], mirror["name"], hostname)
                tmpl_data["check_command"] = "check_nrpe!%s_%s" % (check, mirror["name"])
                tmpl_data["servicegroup"] = conf.nagios_servicegroup()
                checkfile = "service__%s_%s_%s.cfg" % (conf.nagios_hostname(), check, mirror["name"])
                configpath = os.path.join(nagios_exportdir, checkfile)
                tmplfile = "nagios-config.tmpl"
                file_from_template(tmplfile, configpath, tmpl_data)
                wanted_checks.append(checkfile)
    else:
        log("CHARM: No role assigned to {}; not configuring Nagios checks".format(hostname))

    # Clear out any old role checks
    # Making (fairly) sure we only touch files we manage
    role_check_keys = role_checks.keys()
    # And, to make the legacy per-role rsync checks go away:
    role_check_keys.append('check_rsync')
    valid_beginnings = role_check_keys + ["service__{}_{}".format(conf.nagios_hostname(), s)
                                          for s in role_check_keys]
    for d in [nagios_etcdir, nagios_exportdir]:
        enabled_checks = []
        for f in os.listdir(d):
            if f.startswith(tuple(valid_beginnings)):
                enabled_checks.append(f)

        unwanted_checks = list(set(enabled_checks) - set(wanted_checks))
        for check in unwanted_checks:
            try:
                checkpath = os.path.join(d, check)
                log("CHARM: Removing nrpe check {}".format(checkpath))
                os.unlink(checkpath)
            except OSError:
                log("CHARM: Unable to delete nrpe check {}".format(checkpath))

    # Add disk checks
    paths_to_check = [path for path in collect_mountpoints(paths) if path != '/']
    for path in paths_to_check:
        generate_disk_check(conf, path)

    # RT#79518: Add prompt-critical check for all disks; taken from puppet.
    check_all_disks = juju_header()
    check_all_disks += "command[check_all_disks_promptcrit]=/usr/lib/nagios/plugins/check_disk -L -w 2% -c 1% -K 4% -A -i '^/run/lxcfs' -i '^/sys' -i '^/snap'\n"  # noqa: E501
    write_file(os.path.join(nagios_etcdir, 'check_all_disks_promptcrit.cfg'),
               check_all_disks)

    log("CHARM: Reloading nagios-nrpe-server")
    service_reload('nagios-nrpe-server')


def configure_log_archiving(conf, hostname):
    settings = {}
    logdirs = []

    # Create a list of logdirs to archive
    # Store them as a string of base64-encoded words
    # to avoid weird characters upsetting things
    services = ["apache", "ftp", "rsync"]
    for service in filter(conf.archive_logs_for, services):
        logdirs.append(conf.logdir(service))
    settings["logdirs"] = " ".join(map(base64.b64encode, logdirs))
    settings["archive_after"] = conf.archive_logs_after()
    settings["parent_unit_name"] = local_unit()
    if relation_id():
        # In a relation hook
        relation_set(None, settings)
    else:
        # Already established relation
        for r in relation_ids("log-archive"):
            relation_set(r, settings)


def configure_extra_mirrors(conf, hostname):
    settings = {}
    if is_relation_made("ubuntu-mirror"):
        settings["apache_logdir"] = conf.logdir("apache")
        settings["role_map"] = json.dumps(conf.roles())
        settings["cdimage_mirror_map"] = json.dumps(conf.mirror_map("cdimage"))
        settings["releases_mirror_map"] = json.dumps(conf.mirror_map("releases"))
        settings["cloud-images_mirror_map"] = json.dumps(conf.mirror_map("cloud-images"))
        settings["simple-streams_mirror_map"] = json.dumps(conf.mirror_map("simple-streams"))
        settings["ubuntu_mirror_map"] = json.dumps(conf.mirror_map("ubuntu"))
        settings["parent_app_name"] = conf.app_name()
        settings["parent_unit_name"] = local_unit()
    if conf.nagios_relation_ok():
        settings["nagios_servicegroup"] = conf.nagios_servicegroup()
    # Base64 everything
    encoded = {}
    for d in settings:
        encoded[d] = base64.b64encode(settings[d])
    if relation_id():
        log("CHARM: In a relation hook ({})- sending details".format(relation_id()))
        relation_set(None, encoded)
    else:
        log("CHARM: Not in a relation hook - sending details")
        for r in relation_ids("ubuntu-mirror"):
            log("CHARM: Sending details to id: {}".format(r))
            relation_set(r, encoded)


# https://discuss.linuxcontainers.org/t/how-to-detect-lxd-from-inside-container/6042
def is_lxc():
    if os.getuid() != 0:
        raise Exception('is_lxc: cannot detect lxd - not running as root')

    if 'container=lxc' in open('/proc/1/environ').read().split('\0'):
        return True

    return False


def configure_sysctl(conf, hostname):
    # In a container, these keys exist but can't be changed.  sysctl
    # seems to realize that if setting these fail as root, something
    # is beyond our control and exits 0.  Only add such keys here.
    general_settings = {
        # We now run with 32k apache_max_clients, fix issue with thread creation
        # https://publib.boulder.ibm.com/httpserv/ihsdiag/apr_thread_create.html
        'kernel.pid_max': 65536,
        # RT#90964: Kill swap on ubuntu-archive servers
        'vm.swappiness': 0,
        # network-online.target but may not be fully up per upstream bug. Let's use
        # ip_nonlocal_bind for now.
        # https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=868650
        'net.ipv4.ip_nonlocal_bind': 1,
        'net.ipv6.ip_nonlocal_bind': 1,
    }
    sysctl.create(yaml.dump(general_settings), '/etc/sysctl.d/90-ubuntu-mirror.conf')

    # net.core and net.ipv6.route are empty in containers.  sysctl
    # exits 255 for nonexistent keys, so don't even bother trying.
    non_container_settings = {
        # RT#102466: BBR congestion control
        'net.core.default_qdisc': 'fq',
        'net.ipv4.tcp_congestion_control': 'bbr',
        # RT#68166: Tune kernel for larger number of IPv6 users
        'net.ipv6.route.max_size': 16384,
    }
    if not is_lxc():
        sysctl.create(yaml.dump(non_container_settings), '/etc/sysctl.d/90-ubuntu-mirror-non-container.conf')


def configure_user(conf, hostname):
    if not user_exists(conf.mirror_user()):
        mirror_userinfo = adduser(conf.mirror_user(), system_user=True)
        if not os.path.isdir(mirror_userinfo.pw_dir):
            mkdir(mirror_userinfo.pw_dir, owner=conf.mirror_user())


def configure_directories(conf, hostname):
    if not os.path.isdir("/srv/ftp.root"):
        mkdir("/srv/ftp.root")


def configure_keyring(conf, hostname):
    import_cmd = [
        '/usr/bin/gpg', '--no-default-keyring', '--keyring', '/usr/local/share/ubuntu-mirror-charm.gpg', '--import']
    keys = glob(os.path.join(charm_dir(), 'keys', '*.asc'))
    import_cmd.extend(keys)
    check_call(import_cmd)


@hooks.hook("install.real")
def install():
    conf = Config()
    hostname = socket.gethostname()
    log("CHARM: Installing {}".format(conf.app_name()))
    apt_install(required_pkgs, options=['--force-yes'])
    configure_user(conf, hostname)
    configure_scripts(conf, hostname)
    configure_directories(conf, hostname)
    for module in apache_modules:
        check_call(["/usr/sbin/a2enmod", module])


@hooks.hook("config-changed")
def config_changed():
    conf = Config()
    hostname = socket.gethostname()
    log("CHARM: Configuring {}".format(conf.app_name()))
    configure_user(conf, hostname)
    configure_scripts(conf, hostname)
    configure_directories(conf, hostname)
    configure_rsync_client(conf, hostname)
    configure_rsync_server(conf, hostname)
    configure_apache(conf, hostname)
    configure_bind_mounts(conf, hostname)
    configure_vsftp(conf, hostname)
    configure_triggers(conf, hostname)
    configure_nrpe(conf, hostname)
    configure_log_archiving(conf, hostname)
    configure_sysctl(conf, hostname)
    configure_keyring(conf, hostname)


@hooks.hook("upgrade-charm")
def upgrade_charm():
    conf = Config()
    log("CHARM: Upgrading {}".format(conf.app_name()))
    install()


@hooks.hook("nrpe-external-master-relation-changed")
def nrpe_external_master_relation_changed():
    conf = Config()
    hostname = socket.gethostname()
    log("CHARM: Changing nrpe relation for {}".format(conf.app_name()))
    configure_nrpe(conf, hostname)
    configure_extra_mirrors(conf, hostname)


@hooks.hook("local-monitors-relation-changed")
def local_monitors_relation_changed():
    conf = Config()
    hostname = socket.gethostname()
    log("CHARM: Changing monitors relation for {}".format(conf.app_name()))
    configure_nrpe(conf, hostname)


@hooks.hook("log-archive-relation-changed")
def log_archive_relation_changed():
    conf = Config()
    hostname = socket.gethostname()
    log("CHARM: Changing log-archive relation for {}".format(conf.app_name()))
    configure_log_archiving(conf, hostname)


@hooks.hook("ubuntu-mirror-relation-changed")
def ubuntu_mirror_relation_changed():
    conf = Config()
    hostname = socket.gethostname()
    log("CHARM: Changing ubuntu-mirror relation for {}".format(conf.app_name()))
    configure_extra_mirrors(conf, hostname)


if __name__ == "__main__":
    hook_name = os.path.basename(sys.argv[0])
    if hook_name.endswith('.real'):
        hook_name = hook_name[:-5]
    execd_dir = os.path.join(charm_dir(), 'exec.d')
    execd_run(('charm-pre-%s' % hook_name), execd_dir=execd_dir)
    hooks.execute(sys.argv)
    execd_run(('charm-post-%s' % hook_name), execd_dir=execd_dir)
